'''

TODO


Non random starting positions
Turn off bullets on win / lose

Waves
Win / Lose screens
Set waypoints with ctrl
Cycle through ships

Distance circles
Patrol boat
Battleship

Monolithic file alert!

[DONE] HP bars should be billboarded
[DONE] Exit back to menu
[DONE] Remove duplicate attack markers


'''



from direct.showbase.DirectObject import DirectObject
from direct.showbase import PythonUtil

from pandac.PandaModules import *


from controller import TickController


from view import sprite
from view import Effects
from systems import Physics

import random, sys, math
from view import PolyButton

from copy import copy

SHIPNAMES = [
'Thunder Child',
'Charybdis',
'Atermis',
'Ulysses',
'Vagabond',
'Viking',
'Viperous',
'Warlock',
'Avenger',
'Defiant',
'Bounty',
'Venus',
'Indefatigable'
'Surprise',
'Dauntless',
'Providence'
]

class App(DirectObject):

  def __init__(self):

    self.g = None

    self.accept('escape', messenger.send, ['Quit'])

    render.prepareScene(base.win.getGsg())


    self.title()

    self.accept('Quit', self.exit)

  def title(self):

    # Setup PolyButton
    # Very cool snippet to create 3d buttons without dealing with DirectGUI / Librocket
    # http://www.panda3d.org/forums/viewtopic.php?t=2954
    PolyButton.setup( parent=render2d, startNow=True )

    # clears aspect2d's OmniBoundingVolume and its children's bounds test blockade,
    # so collision test performance against aspect2d is at its best.
    aspect2d.node().setFinal(0)
    aspect2d.node().clearBounds()


    # Show the title background
    self.sprite = sprite.Sprite2d('assets/title.png', rows=1, cols=1)
    self.sprite.node.setScale(0.17)
    self.sprite.node.reparentTo(aspect2d) 

    # Start Button
    self.stnp=loader.loadModel('assets/ui/button-start/but')
    self.stnp.reparentTo(aspect2d)
    self.stnp.setTransparency(1)
    self.buttonStart = PolyButton.myButton(self.stnp, 'Cube', ('hover','pressed','disabled'), command=( self.start ))  

    # Exit Button
    self.exnp=loader.loadModel('assets/ui/button-exit/but')
    self.exnp.reparentTo(aspect2d)
    self.exnp.setTransparency(1)
    self.exnp.setZ(self.exnp, -0.2)
    self.buttonExit = PolyButton.myButton(self.exnp, 'Cube', ('hover','pressed','disabled'), command=( self.exit ))  

    # Need to reset the texture filtering so its not blurred
    for tex in self.buttonStart.stateTextures:
      tex.setMagfilter(Texture.FTNearest)
      tex.setMinfilter(Texture.FTNearest)


    

  def start(self):

    self.sprite.clear()
    PolyButton.stop()
    # Remove the buttons
    self.buttonStart.removeButton()
    self.buttonExit.removeButton()
    self.exnp.removeNode()
    self.stnp.removeNode()
    
    self.buttonStart = ''
    self.buttonExit = ''

    self.accept('r-up', self.restart)
    self.accept('q-up', self.stop)

    self.g = Game()

  def restart(self):

    self.stop()
    self.g = Game()


  def stop(self):

    if(self.g):
      self.g.destroy()
      del self.g

  def exit(self):

    if(self.g):
      self.stop()
      PolyButton.reset()
      self.title()
    else:
      sys.exit

class SoundContoller(DirectObject):

  def __init__(self):

    self.hit = base.loader.loadSfx("assets/sounds/Hit_Hurt.wav")
    self.hit.setVolume(0.5)

    self.shoot1 = base.loader.loadSfx("assets/sounds/shoot2.wav")
    self.shoot1.setVolume(0.5)

    self.exp = base.loader.loadSfx("assets/sounds/Explosion.wav")
    #self.exp.setVolume(0.2)

    self.music = base.loader.loadMusic("assets/music/DST-DasElectron.ogg")
    self.music.setLoop(0) 
    self.music.setVolume(0.3)
    self.music.play()

    self.accept('BulletHit', self.playHit)
    self.accept('DestroyShip', self.playExp)
    self.accept('BulletAdded', self.playShoot1)

  def playExp(self, a):
    self.exp.play()

  def playHit(self, a, b):
    self.hit.play()

  def playShoot1(self, a, b):
    self.shoot1.play()


class MouseController(DirectObject):

  def __init__(self):
    self.z = 0
    self.plane = Plane(Vec3(0, 0, 1), Point3(0, 0, self.z))

    # This is for checking the mouse position
    self.model = loader.loadModel('assets/nothing')
    self.model.reparentTo(render)
    self.model.setScale(0.01)
    #self.model.hide()


    self.accept('mouse1', self.mousePressed, [1])
    self.accept('mouse1-up', self.mousePressed, [0])

    self.accept("wheel_up",lambda : self.mouseWheel(10))
    self.accept("wheel_down",lambda : self.mouseWheel(-10))

    self.accept('mouse3', self.mouse2Pressed, [1])
    self.accept('mouse3-up', self.mouse2Pressed, [0])


    self.accept('1-up', messenger.send, ['SelectShip'])

    self.accept('TickEvent', self.__getMousePos)



  def mouseWheel(self, v):

    base.cam.setY(base.cam.getY() + v)

  def mousePressed(self, v):

    messenger.send('CommandMovePlayer', [self.model.getPos()])


  def mouse2Pressed(self, v):

    messenger.send('CommandAttackPlayer', [self.model.getPos()])

  def __getMousePos(self, dt):
    if base.mouseWatcherNode.hasMouse():
      mpos = base.mouseWatcherNode.getMouse()
      pos3d = Point3()
      nearPoint = Point3()
      farPoint = Point3()
      base.camLens.extrude(mpos, nearPoint, farPoint)
      if self.plane.intersectsLine(pos3d,
          render.getRelativePoint(base.cam, nearPoint),
          render.getRelativePoint(base.cam, farPoint)):
        self.model.setPos(render, pos3d)

      # Mouse at screen edge
      panx = 0
      pany = 0

      if(mpos[0] < -0.9):
        panx = -10

      if(mpos[0] > 0.9):
        panx = 10

      if(mpos[1] < -0.9):
        pany = -10

      if(mpos[1] > 0.9):
        pany = 10

      if(panx or pany):
        messenger.send('MouseScreenEdge', [panx * dt, pany * dt])



class Ocean(DirectObject):

  def __init__(self):

    node = loader.loadModel('assets/ocean')

    node.setLightOff()

    node.reparentTo(render)

    tex = loader.loadTexture('assets/ocean01.png')
    tex.setMagfilter(Texture.FTNearest)
    tex.setMinfilter(Texture.FTNearest)
    tex.setAnisotropicDegree(2)


    ts = TextureStage('ts')

    node.setTexture(ts, tex)
    node.setTexScale(ts, 20, 20);
    node.setAttrib(ShadeModelAttrib.make(ShadeModelAttrib.MFlat))
    node.setScale(200)
    node.setTransparency(True)



class Bullet(DirectObject, object):

  def __init__(self, v, pos, human):

    self.node = render.attachNewNode('bullet')
    self.node.setPos(pos)
    self.node.setHpr(v)

    self.name = 'bullet'

    self.human = human

    self.firedFrom = pos

    self.range = 30

class BulletController(DirectObject):

  def __init__(self):

    self.bullets = []

    self.accept('FireBullet', self.add)
    self.accept('BulletHit', self.hit)
    self.accept('BulletDestroy', self.destroy)
    self.accept('TickEvent', self.update)

  def add(self, v, pos, human):

    bullet = Bullet(v, pos, human)

    self.bullets.append(bullet)

    v = render.getRelativeVector(bullet.node,Vec3(0,0,1))

    messenger.send('BulletAdded', [v, bullet])

  def hit(self, e, p):
    if(e.node and p.node):
      vec=e.node.getPos()-p.node.getPos()
      vec.normalize()
      e.setForce(vec)

      e.damage(1)

  def destroy(self, bullet):
    if(bullet in self.bullets):
      self.bullets.remove(bullet)

  def update(self, dt):

    for bullet in self.bullets:

      bullet.node.setPos(bullet.node, Vec3(0,0,1))
      v = bullet.node.getPos() - bullet.firedFrom
      if(v.length() > bullet.range):
        messenger.send('BulletDestroy', [bullet])


class Turret(DirectObject):

  def __init__(self, human):

    self.sprite = sprite.Sprite2d('assets/turret1.png', rows=1, cols=2)
    self.sprite.node.setZ(1.5)
    self.sprite.node.setP(-90)
    self.sprite.node.setLightOff()

    # This is kinda dodgy
    self.human = human

    self.cooldown = 1
    self.coolup = 0

  def fire(self, dt):
    
    self.coolup -= dt
    if(self.coolup <= 0):
      self.sprite.setFrame(1)
      v = render.getRelativeVector(self.sprite.node,Vec3(0,-1,0))
      v *= 1000.0

      dispersion = random.random() * 10

      dis = Vec3(dispersion, 0, 0)

      hpr = self.sprite.node.getHpr(render)

      hpr += dis

      messenger.send('FireBullet', [hpr, self.sprite.node.getPos(render), self.human])

      self.coolup = self.cooldown


    if(self.coolup <= 1):
      self.sprite.setFrame(0)


class Ship(DirectObject, object):

  def __init__(self, pos, player):

    self.hitpoints = 100
    self.name = random.choice(SHIPNAMES)

    self.node = render.attachNewNode('ship')
    self.node.setPos(pos)

    # A push force
    self.force = Vec3(0,0,0)

    self.commandSystem = CommandSystem(self)

    self.player = player

class Debris(DirectObject):

  def __init__(self, pos, hpr, img):

    self.node = render.attachNewNode('debris')
    self.node.setPos(pos)
    self.node.setHpr(hpr)
    self.sprite = sprite.Sprite2d(img, rows=1, cols=1)
    self.sprite.node.setZ(1)
    self.sprite.node.setP(-90)
    self.sprite.node.reparentTo(self.node)

    self.force = Vec3(0,0,0)

  def update(self, dt):

    if(self.node):      

      self.node.setPos(self.node, self.force[0] * dt, self.force[1] * dt, 0)

      self.force *= 0.99

class Destroyer(Ship):

  def __init__(self, pos, player):

    super(Destroyer, self).__init__(pos, player)    

    self.sprite = sprite.Sprite2d('assets/ship2.png', rows=1, cols=2)
    self.sprite.setFrame(0)
    self.sprite.node.setZ(1)
    #self.sprite.node.setX(0.5)
    #self.sprite.node.setY(-2)
    #self.sprite.node.setR(-90)
    self.sprite.node.setP(-90)

    self.sprite.node.reparentTo(self.node)    

    self.turret = Turret(player.humanControlled)
    self.turret.sprite.node.reparentTo(self.node)
    self.turret.sprite.node.setY(2)

    self.turrets = []
    self.turrets.append(self.turret)

    turret = Turret(player.humanControlled)
    turret.sprite.node.reparentTo(self.node)
    turret.sprite.node.setY(-2)
    self.turrets.append(turret)

    self.healthbar = sprite.Sprite2d('assets/hp.png', rows=1, cols=1)
    self.healthbar.node.setZ(1)
    self.healthbar.node.setY(6)
    #self.healthbar.node.setP(-90)
    self.healthbar.node.setSx(50)
    self.healthbar.node.setSz(0.1)
    self.healthbar.node.setBillboardPointEye()
    self.healthbar.node.reparentTo(self.node)    

    self.target = False
    self.attackTarget = False

    self.maxspeed = 10
    self.speed = 0
    self.acceleration = 0.6


  def setForce(self, f):

    self.force = f

  # Turn all weapons to target
  def trackTarget(self, dt, target):
    for t in self.turrets:

       projturn = 180

       # Calculate required heading
       # Should put this somewhere, since we call it in a few places
       oldHpr = copy(t.sprite.node.getHpr() )
       t.sprite.node.lookAt(render, target)
       newHpr = t.sprite.node.getHpr()
       t.sprite.node.setHpr(oldHpr)

       targetH = PythonUtil.fitDestAngle2Src(oldHpr[0], newHpr[0])

       #print(str(oldHpr) + ':' + str(newHpr))

       if targetH - oldHpr[0] > dt*projturn:
          targetH = oldHpr[0] + dt*projturn
       elif targetH - oldHpr[0] < -dt*projturn:
          targetH = oldHpr[0] - dt*projturn

       t.sprite.node.setH(targetH)


  def turnTo(self, dt, target, node):    

    if(node):
      # Calculate required heading
      oldHpr = copy(node.getHpr() )
      node.lookAt(render, target)
      newHpr = node.getHpr()
      node.setHpr(oldHpr)

      targetH = PythonUtil.fitDestAngle2Src(oldHpr[0], newHpr[0])

      projturn = 100 * (abs(newHpr[0] - oldHpr[0]) / 360)

      #print(str(oldHpr) + ':' + str(newHpr))

      if targetH - oldHpr[0] > dt*projturn:
        targetH = oldHpr[0] + dt*projturn
      elif targetH - oldHpr[0] < -dt*projturn:
        targetH = oldHpr[0] - dt*projturn

      node.setH(targetH)


  def update(self, dt):

    if(self.node):

      # Keep the health bar above the ship
      hppos = self.node.getPos(render)
      hppos[1] += 4
      self.healthbar.node.setPos(render, hppos)

      if(self.attackTarget):

        self.trackTarget(dt, self.attackTarget)

        for turret in self.turrets:
          turret.fire(dt)

      if(self.target):

        if(self.speed < self.maxspeed):
          self.speed = (self.speed * self.acceleration) + 0.5
        else:
          self.speed = self.maxspeed

        self.turnTo(dt, self.target, self.node) 

        current2dPos = self.node.getPos(render)
        current2dPos.z = 0

        travelVec = current2dPos - self.target

        if(travelVec.length() < 1.0):
          self.target = False
          messenger.send('CommandMoveComplete', [self])

      else:
        if(self.speed > 0):
          self.speed -= 0.1

      if(self.speed < 0):
        self.speed = 0

      self.node.setPos(self.node, 0, self.speed * dt, 0)

      self.force *= 0.2

      self.node.setPos(self.node, self.force[0], self.force[1], 0)


  def moveTo(self, pos):

    self.target = pos

  def attack(self, pos):

    self.attackTarget = pos

  def damage(self, v):
    messenger.send('shake', [0.1])
    self.hitpoints -= v
    if(self.hitpoints >= 0):

      if(self.hitpoints == 0):
        messenger.send('DestroyShip', [self])
        splode = Effects.GlassExplosion(self.node.getX(), self.node.getY())
        messenger.send('shake', [2])
        self.destroy()
      else:        
        self.healthbar.node.setSx(self.hitpoints / 2)

  def destroy(self):
    self.turrets = []
    self.healthbar.node.removeNode()

    if(self.commandSystem):
      self.commandSystem.ignoreAll()
      self.commandSystem = ''

    self.ignoreAll()

    self.node.removeNode()

class Player(DirectObject):

  def __init__(self, hc = False, name = '', colour = 0):

    self.humanControlled = hc

    self.name = name

    self.fleet = []

    self.colour = colour

  def destroy(self):

    self.fleet = []


class Command(DirectObject):

  def __init__(self, command, v):

    self.command = command
    self.v = v


class CommandSystem(DirectObject):

  def __init__(self, ship):

    self.ship = ship

    self.commands = []
    self.currentCommand = ''

    self.accept('CommandMove', self.moveTo)
    self.accept('CommandMoveComplete', self.commandComplete)
    self.accept('CommandAttack', self.attack)

    self.accept('TickEvent', self.update)

  def update(self, dt):

    if(len(self.commands) == 0):

      test = 1
      #print ('No commands')

    else:

      if(not self.currentCommand):
        self.currentCommand = self.commands[0]
        self.ship.moveTo(self.currentCommand.v)


  def commandComplete(self, ship):
    if(self.currentCommand):
      self.currentCommand = ''
      self.commands.pop(0)

  def moveTo(self, ship, pos):

    if(self.ship == ship):
      if(len(self.commands) < 5):
        self.commands.append(Command('Move', pos))


  def attack(self, ship, pos):

    if(self.ship == ship):

      self.ship.attack(pos)



class Intel(DirectObject):

  def __init__(self, player, human):

    self.player = player

    self.human = human

    self.accept('TickEvent', self.update)

  def update(self, dt):

    for ship in self.player.fleet:

      pos = ship.node.getPos() + (25 - (random.random() * 50), 25 - (random.random() * 50), 0)

      messenger.send('CommandMove', [ship, pos])

      # Attack a human ship
      if(len(self.human.fleet) > 0):
        target = self.human.fleet[0]
        distance = target.node.getPos() - ship.node.getPos()
        if(distance.length() < 25):          
          messenger.send('CommandAttack', [ship, target.node.getPos()])

        

class Game(DirectObject):
  '''
    The main game object / world
  '''
  def __init__(self):

    self.textShips = TextNode('node name')
    self.textShips.setText("Enemy ships left: " + str(0))
    self.textShips.setShadow(0.05, 0.05)
    self.textShips.setShadowColor(0, 0, 0, 1)
    textNodePath = aspect2d.attachNewNode(self.textShips)
    textNodePath.setScale(0.05)
    textNodePath.setX(-1.2)
    textNodePath.setZ(0.90)



    print 'Game started'
    self.gameroot = render.attachNewNode('Gameroot')

    # Setup Camera
    base.camLens.setFar(1000)
    base.camLens.setNear(0.01)
    self.camPivot = self.gameroot.attachNewNode('')
    camera.reparentTo(self.camPivot)
    camera.setPos(0,0,100)
    camera.setP(-90)



    self.tickController = TickController.TickController()
    self.sc = SoundContoller()
    self.mouseController = MouseController()
    self.effectController = Effects.EffectController()

    self.shake = Effects.Shake(base.cam)
    self.physics = Physics.PhysicsSystem(self)
    self.bullets = BulletController()

    self.sprite = sprite.Sprite2d('assets/background.png', rows=1, cols=1)

    self.sprite.node.setX(0)
    self.sprite.node.setY(2)
    self.sprite.node.setZ(-0.0)
    self.sprite.node.setScale(0.02)
    self.sprite.node.setDepthWrite(0)
    self.sprite.node.setDepthTest(0)
    self.sprite.node.setBin('background', 10)
    self.sprite.node.setLightOff()

    self.sprite.node.reparentTo(base.cam)

    self.ocean = Ocean()

    self.entities = []
    self.entitiesGarbage = []

    self.players = []

    self.player = Player(True, 'Michael', 0)

    self.players.append(self.player)

    self.setupFleet(self.player)


    self.addComputerPlayer()


    self.shipSelected = 0

    self.accept('SelectShip', self.selectShip)
    self.accept('MouseScreenEdge', self.panCamera)

    self.accept('DestroyShip', self.destroyShip)

    self.accept('CommandMovePlayer', self.commandMove)
    self.accept('CommandAttackPlayer', self.commandAttack)

    self.accept('TickEvent', self.update)

    self.waypoints = ''
    self.atksprite = ''
    self.computerShipCount = 0
    self.startHuman = Vec2(0,2)

  def destroy(self):

    taskMgr.remove('gameLoop')

    self.aiController.ignoreAll()
    del(self.aiController)

    self.tickController.ignoreAll()
    del(self.tickController)

    self.mouseController.ignoreAll()
    del(self.mouseController)

    self.effectController.ignoreAll()
    del(self.effectController)

    self.shake.ignoreAll()
    del(self.shake)

    self.physics.ignoreAll()
    del(self.physics)

    self.bullets.ignoreAll()
    del(self.bullets)

    for e in self.entities:

      e.destroy()
      del(e)

    for e in self.players:

      e.destroy()
      del(e)

    self.ignoreAll()
    self.gameroot.removeNode()


  def destroyShip(self, ship):

    print 'Destroy ' + ship.name

    debris1 = Debris(ship.node.getPos(), ship.node.getHpr(), 'assets/ship2-debris1.png')

    v = ship.node.getRelativeVector(ship.node,Vec3(0,1,0))
    v *= 40
    v.normalize()
    debris1.force = v

    debris2 = Debris(ship.node.getPos(), ship.node.getHpr(), 'assets/ship2-debris2.png')

    v = ship.node.getRelativeVector(ship.node,Vec3(0,-1,0))
    v *= 40
    v.normalize()
    debris2.force = v

    self.entities.append(debris1)
    self.entities.append(debris2)



    for shippy in self.entities[:]:
      if(ship is shippy):
        self.entities.remove(shippy)
        
    for shippy in ship.player.fleet[:]:
      if(ship is shippy):
        ship.player.fleet.remove(shippy)

    self.textShips.setText("Enemy ships left: " + str(self.computerShipCount))



  def panCamera(self, x,y):

    bounds = 100

    # Restrict to bounds
    if((self.camPivot.getX() + x) > -bounds and (self.camPivot.getX() + x) < bounds):
      self.camPivot.setX(self.camPivot, x)

    if((self.camPivot.getY() + y) > -bounds and (self.camPivot.getY() + y) < bounds):      
      self.camPivot.setY(self.camPivot, y)


  def addComputerPlayer(self):

    player = Player(False, 'Test', 1)
    self.players.append(player)
    self.setupFleet(player, 5)
    self.computerShipCount = 5
    self.textShips.setText("Enemy ships left: " + str(5))
    self.aiController = Intel(player, self.player)

  def drawWaypoints(self):

    if(self.waypoints):

      self.waypoints.removeNode()

    # Do we actually have a fleet?
    if(len(self.player.fleet) > 0):

      ship = self.player.fleet[self.shipSelected]

      if(ship.node):

        currentPos = ship.node.getPos()

        segs = LineSegs()
        segs.setThickness( 1.0 )
        segs.setColor( Vec4(0,0.01,0,1) )

        segs.moveTo( currentPos )

        for command in ship.commandSystem.commands:

          segs.drawTo(command.v)

        self.waypoints = render.attachNewNode(segs.create())
        #self.waypoints.setTransparency(True)
        self.waypoints.setLightOff()
        #self.waypoints.setAlphaScale(0.5)

  def commandAttack(self, pos):

    if(not self.atksprite):
      self.atksprite = sprite.Sprite2d('assets/attack.png', rows=1, cols=1)
      self.atksprite.node.reparentTo(render)

    self.atksprite.node.setPos(pos)
    self.atksprite.node.setZ(1)
    self.atksprite.node.setP(-90)    

    if(len(self.player.fleet) > 0):
      messenger.send('CommandAttack', [self.player.fleet[self.shipSelected], pos])


  def commandMove(self, pos):

    if(len(self.player.fleet) > 0):
      # Reset all commands
      self.player.fleet[self.shipSelected].commandSystem.currentCommand = ''
      self.player.fleet[self.shipSelected].commandSystem.commands = []
      messenger.send('CommandMove', [self.player.fleet[self.shipSelected], pos])


  def update(self, dt):

    for e in self.entitiesGarbage:
      e.destroy()
      del(e)

    # Check if no ships left
    playerShipCount = 0
    computerShipCount = 0  

    # Update all ships
    for ship in self.entities:

      if(isinstance(ship, Ship)):

        if(ship.player.humanControlled):
          playerShipCount += 1
        else:
          computerShipCount += 1

      ship.update(dt)

    if(self.computerShipCount != computerShipCount):
      self.computerShipCount = computerShipCount
      self.textShips.setText("Enemy ships left: " + str(self.computerShipCount))
    
    if(computerShipCount == 0):
      print 'You win!'

    if(playerShipCount == 0):
      print 'You lose!'

    self.drawWaypoints()


  def addToFleet(self, ship, player):

    print('Ship added to fleet')

    # Tint ship by player colour
    # ship.sprite.node.setColor(player.colour[0],player.colour[1],player.colour[2], 1)
    ship.sprite.setFrame(player.colour)

    player.fleet.append(ship)
    messenger.send('ShipAdded', [ship, player])

  def selectShip(self):

    self.shipSelected += 1

    if(self.shipSelected > len(self.player.fleet) - 1):
      self.shipSelected = 0

    print self.shipSelected

    print('Ship selected' + self.player.fleet[self.shipSelected].name)


  def setupFleet(self, player, numOfShips = 1):

    print('Setup fleet')

    for x in range(0, numOfShips):

      if(player.humanControlled):

        posx = 0
        posy = 0

      else:

        posx = random.random() * 50
        posy = 105 - random.random() * 5

      ship = Destroyer((posx, posy, 0), player)

      self.entities.append(ship)
      self.addToFleet(ship, player)