'''

'''
from direct.showbase.DirectObject import DirectObject
from pandac.PandaModules import *

from panda3d.core import Vec3
from panda3d.core import Vec4
from panda3d.core import BitMask32
from panda3d.bullet import *

maskFriendly = BitMask32.bit(0)
maskEnemy = BitMask32.bit(1)
maskFriendlyBullet = BitMask32.bit(2)
maskEnemyBullet = BitMask32.bit(2)

class PhysicsSystem(DirectObject):

  def __init__(self, world):

    self.world = world

    self.worldNP = render.attachNewNode('World')

    # World
    self.bulletWorld = BulletWorld()
    self.bulletWorld.setGravity(Vec3(0, 0, -9.81))

    self.setupCollisionMasks(self.bulletWorld)

    self.destroyQueue = []

    # Setup Debug node
    self.debugNP = self.worldNP.attachNewNode(BulletDebugNode('Debug'))
    #self.debugNP.show()
    self.debugNP.node().showWireframe(True)
    self.debugNP.node().showConstraints(False)
    self.debugNP.node().showBoundingBoxes(False)
    self.debugNP.node().showNormals(False)  
    self.bulletWorld.setDebugNode(self.debugNP.node())

    self.accept('BulletAdded', self.addBullet)
    self.accept('BulletDestroy', self.queueDestroy)
    self.accept('ShipAdded', self.addShip)
    self.accept('DestroyShip', self.destroyShip)
    self.accept('TickEvent', self.update)
    self.accept('bullet-contact-added', self.handleCollision)

  def update(self, dt):

    for e in self.destroyQueue:

      self.destroyBullet(e)
      


    self.bulletWorld.doPhysics(dt, 10, 1.0/180.0)

  def addShip(self, ship, player):

    shape = BulletBoxShape(Vec3(1, 5, 10.5))
    body = BulletRigidBodyNode(str(ship))

    bodyNP = render.attachNewNode(body)
    bodyNP.reparentTo(ship.node)
    bodyNP.node().addShape(shape)
    bodyNP.node().setPythonTag('node', ship)
    bodyNP.node().setMass(100)
    bodyNP.node().setKinematic(True)

    # Notify whe a collision has occurred
    #bodyNP.node().notifyCollisions(True)

    if(player.humanControlled):
      bodyNP.setCollideMask(BitMask32.bit(0))
    else:
      bodyNP.setCollideMask(BitMask32.bit(1))

    self.bulletWorld.attachRigidBody(bodyNP.node())
    bodyNP.node().setGravity(Vec3(0,0,0))
    ship.hitbox = bodyNP

  def addBullet(self, v, bullet):

    shape = BulletBoxShape(Vec3(0.05, 0.05, 0.05))
    body = BulletRigidBodyNode(str(bullet))

    bodyNP = render.attachNewNode(body)
    bodyNP.reparentTo(bullet.node)
    bodyNP.node().addShape(shape)
    bodyNP.node().setPythonTag('node', bullet)
    bodyNP.node().setMass(1)
    bodyNP.node().setLinearVelocity(v)
    bodyNP.node().setKinematic(True)

    # Notify whe a collision has occurred
    bodyNP.node().notifyCollisions(True)

    if(bullet.human):
      bodyNP.setCollideMask(BitMask32.bit(2))
    else:
      bodyNP.setCollideMask(BitMask32.bit(3))

    #bodyNP.setCollideMask(BitMask32.bit(1))



    self.bulletWorld.attachRigidBody(bodyNP.node())
    bodyNP.node().setGravity(Vec3(0,0,0))
    bullet.hitbox = bodyNP

  def queueDestroy(self, e):

    self.destroyQueue.append(e)

  def destroyShip(self, ship):

    if(ship.hitbox):
      ship.hitbox.node().clearPythonTag('node')

      self.bulletWorld.removeRigidBody(ship.hitbox.node())
      ship.hitbox.removeNode()

      ship.hitbox = None

  def destroyBullet(self, bullet):

    if(bullet.hitbox):
      bullet.hitbox.node().clearPythonTag('node')

      self.bulletWorld.removeRigidBody(bullet.hitbox.node())
      bullet.hitbox.removeNode()

      bullet.hitbox = None

  def setupCollisionMasks(self, btWorld):
   
    #first disable objects collision with objects from same group (ship-ship, bullet-bullet)
    for i in range(32):
        btWorld.setGroupCollisionFlag(i, i, False)
   
    groups = {"Player":0,
             "Enemy" :1,
             "PlayerBullet":2,
             "EnemyBullet":3,
             "FloorPlane":4
             }

    colliders = {
                 "Player":["Enemy","EnemyBullet"],
                 "Enemy":["Player","PlayerBullet"]
                 }

    for obj,into in colliders.iteritems():
        for val in into:
            btWorld.setGroupCollisionFlag(groups[obj], groups[val], True)

  def handleCollision(self, node1, node2):

    entity1 = node1.getPythonTag('node')
    entity2 = node2.getPythonTag('node')

    if(type(entity2).__name__ == 'Destroyer'):
      messenger.send('BulletHit', [entity2, entity1])

    if(type(entity1).__name__ == 'Destroyer'):
      messenger.send('BulletHit', [entity1, entity2])

    if(type(entity2).__name__ == 'Bullet'):
      messenger.send('BulletDestroy', [entity2])

    if(type(entity1).__name__ == 'Bullet'):
      messenger.send('BulletDestroy', [entity1])

