"""


"""
from direct.showbase.DirectObject import DirectObject
from pandac.PandaModules import *
from direct.particles.ParticleEffect import ParticleEffect
from direct.interval.IntervalGlobal import *

class Shake(DirectObject):

    def __init__(self, node):

        self.x = 0.5
        self.y = 0.5

        self.flip = 1
        self.shaking = 0

        self.node = node

        self.origPos = self.node.getPos()

        self.accept('shake', self.shake)
        self.accept('TickEvent', self.update)

    def shake(self, v):

        self.x = v
        self.y = v

        self.shaking = 1

    def update(self, dt):

        if(self.shaking):

            self.node.setX(self.origPos[0] + (self.x * self.flip))
            self.node.setZ(self.origPos[2] + (self.y * self.flip))

            self.flip = -self.flip

            self.x *= 0.8
            self.y *= 0.8

            if(self.x < 0.01):
                self.shaking = 0

class GlassExplosion(object):

  def __init__(self, x, y):

    np = NodePath("boom" + str(x) + str(y))
    p = ParticleEffect()
    np.reparentTo(render)
    np.setDepthWrite(False)
    np.setShaderOff(1)      
    np.setScale(2)
    np.setPos(x,y,0)

    p.loadConfig('data/particle0.ptf')

    p.start(parent=np)

    Sequence(
           Wait(0.11),
           Func(p.softStop),
           Wait(1.0),
           Func(p.cleanup)
           ).start()



class EffectController(DirectObject):
    
    def __init__(self):

        self.effectList = []
        self.particleList = []
        self.particles = []
        self.tracers = []
        self.hits = []
        self.explosions = []
        self.removals = []

        # Explosion mesh drawer
        self.expmd = MeshDrawer()
        self.expmd.setBudget(1000)
        self.mdexpgennode = self.expmd.getRoot()
        self.mdexpgennode.reparentTo(render)

        exptex = loader.loadTexture("assets/explode1.png")
        exptex.setMagfilter(Texture.FTNearest)
        exptex.setMinfilter(Texture.FTNearest)
        ts = TextureStage('ts')
        ts.setMode(TextureStage.MModulate)

        self.mdexpgennode.setTexture(ts,exptex)        
        self.mdexpgennode.setTransparency(True)
        self.mdexpgennode.setPos(0,0,5)
        self.mdexpgennode.setDepthWrite(False)
        self.mdexpgennode.setTwoSided(True)
        self.mdexpgennode.setLightOff(True)
        self.mdexpgennode.setBin("fixed", 100)

        self.mdexpgennode.node().setBounds(BoundingSphere((0, 0, 0), 500))
        self.mdexpgennode.node().setFinal(True)


        # Hit mesh drawer
        self.hitmd = MeshDrawer()
        self.hitmd.setBudget(2000)
        self.mdhitgennode = self.hitmd.getRoot()
        self.mdhitgennode.reparentTo(render)

        hittex = loader.loadTexture("assets/hit.png")
        hittex.setMagfilter(Texture.FTNearest)
        hittex.setMinfilter(Texture.FTNearest)
        ts = TextureStage('ts')
        ts.setMode(TextureStage.MModulate)

        self.mdhitgennode.setTexture(ts,hittex)        
        self.mdhitgennode.setTransparency(True)
        self.mdhitgennode.setPos(0,0,5)
        self.mdhitgennode.setDepthWrite(False)
        self.mdhitgennode.setTwoSided(True)
        self.mdhitgennode.setLightOff(True)
        self.mdhitgennode.setBin("fixed", 100)

        self.mdhitgennode.node().setBounds(BoundingSphere((0, 0, 0), 500))
        self.mdhitgennode.node().setFinal(True)
        
        # Tracer mesh drawer
        self.tracermd = MeshDrawer()
        self.tracermd.setBudget(2000)
        self.mdtracergennode = self.tracermd.getRoot()
        self.mdtracergennode.reparentTo(render)

        tracertex = loader.loadTexture("assets/bullet.png")
        tracertex.setMagfilter(Texture.FTNearest)
        tracertex.setMinfilter(Texture.FTNearest)
        ts = TextureStage('ts')
        ts.setMode(TextureStage.MModulate)

        self.mdtracergennode.setTexture(ts,tracertex)        
        self.mdtracergennode.setTransparency(True)
        self.mdtracergennode.setPos(0,0,1)
        self.mdtracergennode.setDepthWrite(False)
        self.mdtracergennode.setTwoSided(True)
        self.mdtracergennode.setLightOff(True)
        self.mdtracergennode.setBin("fixed", 100)

        self.mdtracergennode.node().setBounds(BoundingSphere((0, 0, 0), 500))
        self.mdtracergennode.node().setFinal(True)

        self.accept('DestroyShip', self.addExplosion)
        self.accept('BulletAdded', self.addTracer)
        self.accept('BulletDestroy', self.destroyTracer)
        self.accept('TickEvent', self.update)
        self.accept('BulletHit', self.addHit)

    def destroyTracer(self, bullet):
        dirty = ''
        for tracer in self.tracers:

            if(tracer.np == bullet.node):
                dirty = tracer

        if(dirty):
            self.tracers.remove(dirty)

    def addExplosion(self, e):
        np = e.node
        p = np.getPos(render)
        t = Flash(np, p, p, Vec4(0, 0, 1, 1), 0.3, Vec4(1, 1, 1, 1), 25)
        self.explosions.append(t)     

    def addHit(self, e, p):
        np = e.node
        p = p.node.getPos(render)
        t = Flash(np, p, p, Vec4(0, 0, 1, 1), 0.3, Vec4(1, 1, 1, 1), 5)
        self.hits.append(t)

    def addTracer(self, v, bullet):
        np = bullet.node
        v.normalize()
        t = Tracer(np, np.getPos(render), np.getPos(render) + v, Vec4(0, 0, 1, 1), 0.1, Vec4(1, 1, 1, 1))
        self.tracers.append(t)
       
    def update(self, dt):

        # Render explosions
        self.expmd.begin(base.cam, render) 

        for e in self.explosions:

            self.expmd.particle(Vec3(e.start), e.frame, e.scale, e.colour, 0)

            e.life -= 1
            e.scale += 0.15
            if(e.life <= 0):
                
                self.explosions.remove(e)
                del(e)


        self.expmd.end()


        # Render tracers
        self.tracermd.begin(base.cam, render) 

        for tracer in self.tracers:

            pos = tracer.np.getPos()

            v = render.getRelativeVector(tracer.np,Vec3(0,0,1))
            v.normalize()

            self.tracermd.crossSegment(Vec3(pos), Vec3(pos + v), tracer.frame, tracer.scale, tracer.colour)

        self.tracermd.end()

        # Render hits
        self.hitmd.begin(base.cam, render) 

        for hit in self.hits:

            pos = hit.start

            self.hitmd.particle(Vec3(pos), hit.frame, 1, hit.colour, 0)

            hit.life -= 1

            if(hit.life <= 0):
                
                self.hits.remove(hit)
                del(hit)

        self.hitmd.end()




class Flash(object):

    def __init__(self, np, start, stop, frame, scale, colour, life):

        self.np = np
        self.start = start
        self.stop = stop
        self.frame = frame
        self.scale = scale
        self.colour = colour
        self.life = life

class Tracer(object):

    def __init__(self, np, start, stop, frame, scale, colour):

        self.np = np
        self.start = start
        self.stop = stop
        self.frame = frame
        self.scale = scale
        self.colour = colour